import logging
from pathlib import Path
from typing import Union

from abcfold.chai1.af3_to_chai import ChaiFasta
from abcfold.output.file_handlers import (CifFile, ConfidenceJsonFile,
                                          FileTypes, NpyFile, NpzFile)
from abcfold.output.utils import Af3Pae

logger = logging.getLogger("logger")


class ChaiOutput:
    def __init__(
        self,
        chai_output_dir: Union[str, Path],
        input_params: dict,
        name: str,
    ):
        """
        Object to process the output of an Chai-1 run

        Args:
            chai_output_dir (Union[str, Path]): Path to the Chai-1 output directory
            input_params (dict): Dictionary containing the input parameters used for the
            Chai-1 run
            name (str): Name given to the Chai-1 run

        Attributes:
            input_params (dict): Dictionary containing the input parameters used for the
            Chai-1 run
            output_dir (Path): Path to the Chai-1 output directory
            name (str): Name given to the Chai-1 run
            output (dict): Dictionary containing the processed output the contents
            of the Chai-1 output directory. The dictionary is structured as follows:

            {
                1: {
                    "pae": NpzFile,
                    "cif": CifFile,
                    "scores": NpyFile
                },
                2: {
                    "pae": NpzFile,
                    "cif": CifFile,
                    "scores": NpyFile
                },
                ...
            }
            pae_files (list): Ordered list of NpzFile objects containing the PAE data
            cif_files (list): Ordered list of CifFile objects containing the CIF data
            scores_files (list):  Ordered list of NpyFile objects containing the scores
            data

        """
        self.input_params = input_params
        self.output_dir = Path(chai_output_dir)
        self.name = name

        if not self.output_dir.name.startswith("chai1_" + self.name):
            self.output_dir = self.output_dir.rename(
                self.output_dir.parent / f"chai1_{self.name}"
            )
        self.input_fasta = self.get_input_fasta()

        self.output = self.process_chai_output()
        self.pae_files = [
            value["pae"] for value in self.output.values() if "pae" in value
        ]
        self.cif_files = [
            value["cif"] for value in self.output.values() if "cif" in value
        ]
        self.pae_to_af3()
        self.scores_files = [
            value["scores"] for value in self.output.values() if "scores" in value
        ]
        self.af3_pae_files = [
            value["af3_pae"] for value in self.output.values() if "af3_pae" in value
        ]

    def process_chai_output(self):
        file_groups = {}

        for pathway in self.output_dir.iterdir():
            number = pathway.stem.split("model_idx_")[-1]
            if number.isdigit():
                number = int(number)

            file_type = pathway.suffix[1:]

            if file_type == FileTypes.NPZ.value:
                file_ = NpzFile(str(pathway))

            elif file_type == FileTypes.CIF.value:
                file_ = CifFile(str(pathway), self.input_params)
                file_ = self.update_chain_labels(file_)

            elif file_type == FileTypes.NPY.value:
                file_ = NpyFile(str(pathway))
            else:
                continue

            if isinstance(number, str):
                number = -1

            if number not in file_groups:
                file_groups[number] = [file_]
            else:
                file_groups[number].append(file_)

        model_number_file_type_file = {}
        for model_number, files in file_groups.items():
            intermediate_dict = {}
            for file_ in sorted(files, key=lambda x: x.suffix):
                if file_.pathway.stem.startswith("scores.model"):
                    intermediate_dict["scores"] = file_
                elif file_.pathway.stem.startswith("pred.model"):
                    file_.name = f"Chai-1_{model_number}"
                    # Chai cif not recognised by pae-viewer, so we load and save
                    file_.to_file(file_.pathway)
                    intermediate_dict["cif"] = file_
                elif file_.pathway.stem.startswith("pae_scores"):
                    intermediate_dict["pae"] = file_

            model_number_file_type_file[model_number] = intermediate_dict

        model_number_file_type_file = {
            model_number: model_number_file_type_file[model_number]
            for model_number in sorted(model_number_file_type_file)
        }

        return model_number_file_type_file

    def pae_to_af3(self) -> None:
        """
        Convert the Chai-1 PAE data to the format expected by AlphaFold3

        """

        pae_file = self.pae_files[-1]
        for i, cif_file in enumerate(self.cif_files):
            pae = Af3Pae.from_chai1(
                pae_file.data[i],
                cif_file,
            )

            out_name = self.output_dir.joinpath(cif_file.pathway.stem + "_af3_pae.json")
            pae.to_file(out_name)

            self.output[i]["af3_pae"] = ConfidenceJsonFile(out_name)

    def get_input_fasta(self) -> ChaiFasta:
        """
        Function to get the input fasta file used for the Chai-1 run

        Returns:
            ChaiFasta: ChaiFasta object containing the input fasta file

        """

        ch = ChaiFasta(self.output_dir, create_files=False)
        ch.json_to_fasta(self.input_params)

        return ch

    def update_chain_labels(self, cif_file: CifFile) -> CifFile:
        """
        Function to update the chain labels in the CIF file

        Args:
            cif_file (CifFile): CifFile object to update the chain labels for

        """

        cif_file.relabel_chains(self.input_fasta.chain_ids)
        cif_file.to_file(cif_file.pathway)
        return CifFile(cif_file.pathway, self.input_params)



import json
import logging
import re
import warnings
from abc import ABC
from enum import Enum
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

import numpy as np
from Bio.PDB import MMCIFIO, Chain, MMCIFParser, Model
from Bio.PDB.Atom import Atom
from Bio.PDB.kdtrees import KDTree
from Bio.PDB.Polypeptide import is_aa
from Bio.PDB.Residue import Residue
from Bio.SeqUtils import seq1

from abcfold.output.atoms import VANDERWALLS

warnings.filterwarnings("ignore")

logger = logging.getLogger("logger")


class FileTypes(Enum):
    """
    Enum class for the different file types
    """

    NPZ = "npz"
    NPY = "npy"
    CIF = "cif"
    JSON = "json"

    @classmethod
    def values(cls):
        return [value.value for value in cls.__members__.values()]


class ModelCount(Enum):
    """
    Enum class for the different model count types
    """

    ALL = "all"
    RESIDUES = "residues"

    @classmethod
    def values(cls):
        return [value.value for value in cls.__members__.values()]


class ResidueCountType(Enum):
    """
    Enum class for the different residue count types
    """

    AVERAGE = "average"
    CARBONALPHA = "carbonalpha"

    @classmethod
    def values(cls):
        return [value.value for value in cls.__members__.values()]


class FileBase(ABC):
    """
    Abstract base class for the different file types
    """

    def __init__(self, pathway: Union[str, Path]):
        self.pathway = Path(pathway)
        self.suffix = self.pathway.suffix[1:]

    def __str__(self):
        return str(self.pathway)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.pathway})"


class NpzFile(FileBase):

    def __init__(self, npz_file: Union[str, Path]):
        """
        Object to handle npz files

        Args:
            npz_file (Union[str, Path]): Path to the npz file

        Attributes:
            npz_file (Path): Path to the npz file
            data (dict): Dictionary containing the data from the npz file

        """
        super().__init__(npz_file)
        self.npz_file = Path(npz_file)
        self.data = self.load_npz_file()

    def load_npz_file(self) -> dict:
        return dict(np.load(self.npz_file))


class NpyFile(FileBase):
    def __init__(self, npy_file: Union[str, Path]):
        """
        Object to handle npy files

        Args:
            npy_file (Union[str, Path]): Path to the npy file

        Attributes:
            npy_file (Path): Path to the npy file
            data (np.ndarray): Numpy array containing the data from the np
        """

        super().__init__(npy_file)
        self.npy_file = Path(npy_file)
        self.data = self.load_npy_file()

    def load_npy_file(self) -> np.ndarray:
        return np.load(self.npy_file)


class CifFile(FileBase):

    def __init__(self, cif_file: Union[str, Path], input_params: Optional[dict] = None):
        """
        Object to handle cif files

        Args:
            cif_file (Union[str, Path]): Path to the cif file
            input_params (Optional[dict]): Dictionary containing the input parameters
            used for the model. This is used to distinguish between ligands and
            sequences

        Attributes:
            cif_file (Path): Path to the cif file
            model (Structure): BioPython structure object containing the model
            atom_plddt_per_chain (dict): Dictionary containing the pLDDT scores for
            each atom
            residue_plddt_per_chain (dict): Dictionary containing the pLDDT scores for
            each residue
            plddts (list): List containing the pLDDT scores for each atom
            residue_plddts (list): List containing the pLDDT scores for each residue
            name (str): Name given to the model
        """
        if input_params is None:
            self.input_params = {}
        else:
            self.input_params = input_params

        super().__init__(cif_file)
        self.cif_file = Path(cif_file)
        self.clashes = 0
        self.clashes_residues = 0
        self.model = self.load_cif_file()
        self.__ligand_plddts = None
        self.__plddts = None
        self.__residue_plddts = None
        self.__h_score = None
        self.__name = self.cif_file.stem

    @property
    def name(self):
        return self.__name

    @name.setter
    def name(self, name: str):
        if not isinstance(name, str):
            logger.error("Name must be a string")
            raise ValueError()
        self.__name = name

    @property
    def plddts(self):
        """
        The pLDDT scores for each atom in the model
        """
        self.__plddts = [
            plddts for plddts in self.get_plddt_per_atom().values() for plddts in plddts
        ]
        return self.__plddts

    @property
    def residue_plddts(self):
        """
        The pLDDT scores for each residue in the model
        """
        self.__residue_plddts = [
            plddts
            for plddts in self.get_plddt_per_residue().values()
            for plddts in plddts
        ]
        return self.__residue_plddts

    @property
    def average_plddt(self):
        """
        The average pLDDT score for the model
        """
        return float(np.mean(self.plddts))

    @property
    def ligand_plddts(self):
        """
        The pLDDT scores for each ligand in the model
        """
        self.__ligand_plddts = self.get_plddt_per_ligand()
        return self.__ligand_plddts

    @property
    def h_score(self):
        """
        The H score for the model
        """
        self.__h_score = self.calculate_h_score()
        return self.__h_score

    def load_cif_file(self):
        """
        Load the cif file using BioPython
        """
        parser = MMCIFParser(QUIET=True)
        return parser.get_structure(self.pathway.stem, self.pathway)

    def get_chains(self):
        return self.model[0]

    def chain_lengths(
        self,
        mode=ModelCount.RESIDUES,
        ligand_atoms=False,
        ptm_atoms=False,
    ) -> dict:
        """
        Function to get the length of each chain in the model

        Args:
            mode (ModelCount): Enum class specifying the mode to use
            Note: For ligands the length will always be the number of atoms

        Returns:
            dict: Dictionary containing the chain id and the length of the chain

        Raises:
            ValueError: If the mode is not valid
        """
        chains = self.get_chains()
        if mode == ModelCount.ALL or mode == ModelCount.ALL.value:

            # return {
            #     chain.id: len([atom for resiude in chain for atom in resiude])
            #     for chain in chains
            # }
            chain_lengths: dict = {}
            for chain in chains:
                if chain.id in chain_lengths:
                    chain_lengths[chain.id] += len(
                        [atom for resiude in chain for atom in resiude]
                    )
                else:
                    chain_lengths[chain.id] = len(
                        [atom for resiude in chain for atom in resiude]
                    )

            return chain_lengths

        elif mode == ModelCount.RESIDUES or mode == ModelCount.RESIDUES.value:
            residue_counts: dict = {}
            for chain in chains:
                if self.check_other(chain, ["protein", "rna", "dna"]) and ptm_atoms:
                    counter = 0
                    for residue in chain:
                        if is_aa(residue.resname, standard=True):
                            counter += 1
                            continue
                        else:
                            counter += len([atom for atom in residue])

                    residue_counts[chain.id] = counter

                elif self.check_ligand(chain):
                    if ligand_atoms:
                        if chain.id in residue_counts:
                            residue_counts[chain.id] += len(
                                [atom for resiude in chain for atom in resiude]
                            )
                        else:
                            residue_counts[chain.id] = len(
                                [atom for resiude in chain for atom in resiude]
                            )

                    else:
                        residue_counts[chain.id] = 1
                else:
                    residue_counts[chain.id] = len([residue for residue in chain])

            return residue_counts

        else:
            msg = f"Invalid mode. Please use {', '.join(ModelCount.__members__)}"
            logger.critical(msg)
            raise ValueError()

    def token_residue_ids(self) -> dict:
        """
        Function to get the residue ids for each chain in the model

        Returns:
            dict: Dictionary containing the chain id and the residue ids for each
            chain
        """
        from abcfold.output.utils import flatten

        chains = self.get_chains()
        residue_ids = {}
        for chain in chains:
            if self.check_ligand(chain):
                residue_ids[chain.id] = [
                    [residue.id[1]] for residue in chain for _ in residue
                ]
                continue
            residue_ids[chain.id] = [
                (
                    [residue.id[1]]
                    if residue.id[0] == " " or residue.id[0] == "H"
                    else [residue.id[1] for _ in residue]
                )
                for residue in chain
            ]

        residue_ids = {k: flatten(v) for k, v in residue_ids.items()}

        return residue_ids

    def calculate_h_score(self):
        """
        Calculate the H score for the model

        Returns:
            float: The H score for the model
        """

        score = 0
        for i in reversed(range(1, 101)):
            if (100.0 / len(self.plddts)) * np.sum(np.array(self.plddts) >= i) >= i:
                score = i
                break
        return score

    def get_model_sequence_data(self) -> dict:
        """
        Get the sequence for each chain and ligand in the model, used internally
        for plotting

        Returns:
            dict : Chain ID and sequence data
        """
        sequence_data = {}
        for chain in self.model[0]:
            if self.check_ligand(chain):
                sequence_data[chain.id] = "".join(
                    [atom.id[0] for residue in chain for atom in residue]
                )
            else:
                sequence_data[chain.id] = "".join(
                    [seq1(residue.get_resname()) for residue in chain]
                )
        return sequence_data

    def get_plddt_per_atom(self) -> dict:
        """
        Get the pLDDT scores for each atom in the model

        Returns:
            dict: Dictionary containing the chain id and the pLDDT scores for each atom
        """
        plddt: Dict[str, list] = {}
        for chain in self.model[0]:

            for residue in chain:
                for atom in residue:
                    if chain.id in plddt:
                        plddt[chain.id].append(atom.bfactor)
                    else:
                        plddt[chain.id] = [atom.bfactor]

        return plddt

    def get_plddt_per_residue(self, method=ResidueCountType.CARBONALPHA.value) -> dict:
        """
        Get the pLDDT scores for each residue in the model

        Args:
            method (ResidueCountType): Enum class specifying the method to use

        Returns:
            dict: Dictionary containing the chain id and the pLDDT scores for each
            residue
        """
        plddts: Dict[str, list] = {}

        if method not in ResidueCountType.values():
            logger.error(
                f"Invalid method. Please use {', '.join(ResidueCountType.__members__)}"
            )
            raise ValueError()

        chains = self.get_chains()
        for chain in chains:
            if self.check_ligand(chain):
                if chain.id in plddts:
                    plddts[chain.id].extend(
                        [atom.bfactor for residue in chain for atom in residue]
                    )
                else:
                    plddts[chain.id] = [
                        atom.bfactor for residue in chain for atom in residue
                    ]

            else:
                for residue in chain:
                    if method == ResidueCountType.AVERAGE.value:
                        scores = 0
                        for atom in residue:
                            scores += atom.bfactor
                        score = scores / len(residue)

                    elif method == ResidueCountType.CARBONALPHA.value:
                        for atom in residue:
                            if atom.id == "CA":
                                score = atom.bfactor
                                break

                    if chain.id in plddts:
                        plddts[chain.id].append(score)

                    else:
                        plddts[chain.id] = [score]

        plddt_lengths = {k: len(v) for (k, v) in plddts.items()}
        chain_lengths = self.chain_lengths(mode="residues", ligand_atoms=True)
        for chain_id in plddt_lengths:
            assert (
                chain_lengths[chain_id] == plddt_lengths[chain_id]
            ), f"{chain_id}, {chain_lengths[chain_id]} != {plddt_lengths[chain_id]}"
        return plddts

    def get_plddt_per_ligand(self) -> dict:
        """
        Get the pLDDT scores for each ligand in the model

        Returns:
            dict: Dictionary containing the chain id and the pLDDT scores for each atom
        """
        plddt: Dict[str, list] = {}
        for chain in self.model[0]:
            if self.check_ligand(chain):
                for residue in chain:
                    for atom in residue:
                        if chain.id in plddt:
                            plddt[chain.id].append(atom.bfactor)
                        else:
                            plddt[chain.id] = [atom.bfactor]
        return plddt

    def check_ligand(self, chain: Chain) -> bool:
        """
        Check if the chain is a ligand

        Args:
            chain (Chain): BioPython chain object

        Returns:
            bool: True if the chain is a ligand, False otherwise
        """

        return self.check_other(chain, ["ligand"])

    def check_other(self, chain: Chain, check_list) -> bool:
        sequences = self.input_params.get("sequences")
        if sequences is None:
            logger.warning("Unable to gain sequence infromation from input file")
            return False
        for sequence in sequences:
            for sequence_type, sequence_data in sequence.items():
                if sequence_type in check_list:
                    if "id" not in sequence_data:
                        continue
                    if hasattr(chain, "id"):
                        chain_id = chain.id
                    else:
                        chain_id = chain
                    if isinstance(sequence_data["id"], str):
                        if chain_id == sequence_data["id"]:
                            return True
                    elif isinstance(sequence_data["id"], list):
                        if chain_id in sequence_data["id"]:
                            return True
        return False

    def relabel_chains(
        self, chain_ids: List[str], link_ids: Optional[dict] = None
    ) -> None:
        """
        Relabel the chains in the model

        Args:
            chain_ids (List[str]): List of chain ids to relabel the chains

        Returns:
            None
        """

        chain_ids = chain_ids.copy()
        structure = self.model[0]
        old_new_chain_id = {}

        if link_ids is None:
            link_ids = {}
        else:
            for new_ids in link_ids.values():
                for new_id in new_ids:
                    chain_ids.pop(chain_ids.index(new_id))

        old_chain_label_counter, new_chain_label_counter = 0, 0

        chain_names = [chain.id for chain in structure]
        while old_chain_label_counter < len(structure):
            chain = chain_ids[new_chain_label_counter]
            old_new_chain_id[chain_names[old_chain_label_counter]] = chain

            # increment the old_chain everytime a chain has been relabelled
            old_chain_label_counter += 1

            if chain in link_ids:
                ligand_no_added = 2
                for _ in link_ids[chain]:
                    old_new_chain_id[chain_names[old_chain_label_counter]] = chain
                    for residue in structure[chain_names[old_chain_label_counter]]:

                        residue.id = (residue.id[0], ligand_no_added, residue.id[2])
                        ligand_no_added += 1
                    old_chain_label_counter += 1

            new_chain_label_counter += 1

        for chain_to_rename in structure:
            chain_to_rename.id = old_new_chain_id[chain_to_rename.id]

        assert old_chain_label_counter == len(
            self.get_chains()
        ), "Number of chain ids must match the number of chains"
        self.update()

    def update(self):
        self.to_file(self.pathway)
        self = CifFile(self.pathway, self.input_params)

    def reorder_chains(self, new_chain_ids: List[str]):

        assert sorted([chain.id for chain in self.get_chains()]) == sorted(
            new_chain_ids
        ), "The chain ids need to be identical to what is in the model already \
for reordering"

        new_model = Model.Model(0)

        [new_model.add(self.model[0][ch]) for ch in new_chain_ids]
        self.model.detach_child(0)
        self.model.add(new_model)
        self.update()

    def check_clashes(
        self,
        threshold: Union[int, float] = 3.4,
        bucket: int = 10,
        clash_cutoff: float = 0.63,
    ) -> Tuple[List[Tuple[Atom, Atom]], List[Tuple[Residue, Residue]]]:
        """
        Check for clashes between atoms in different chains

        Args:
            threshold: The distance threshold for a clash.

        Returns:
            A list of clashes.

        """
        atoms = self.get_atoms()
        coords = np.array(
            [atom.get_coord() for atom in atoms],
            dtype="d",
        )
        assert bucket > 1
        assert coords.shape[1] == 3
        assert clash_cutoff > 0.0 and clash_cutoff <= 1.0

        tree = KDTree(coords, bucket)
        neighbors = tree.neighbor_search(threshold)
        clashes_atoms, clashes_residues = [], []

        for neighbor in neighbors:
            i1, i2 = neighbor.index1, neighbor.index2
            atom1, atom2 = atoms[i1], atoms[i2]
            # get the element of the atom
            element1 = atom1.element
            element2 = atom2.element
            # find chain_id and residue_id
            chain_id1 = atom1.get_full_id()[2]
            chain_id2 = atom2.get_full_id()[2]

            if chain_id1 == chain_id2:
                continue

            distance = np.linalg.norm(atom1.get_coord() - atom2.get_coord())
            if (atom1.name == "C" and atom1.name == "N") or (
                atom2.name == "N" and atom1.name == "C"
            ):
                continue
            elif (atom1.name == "SG" and atom2.name == "SG") and distance > 1.88:
                continue

            clash_radius = (
                VANDERWALLS.get(element1, 1.7) + VANDERWALLS.get(element2, 1.7)
            ) * 0.63
            if distance < clash_radius:
                residue1 = atom1.get_parent()
                residue2 = atom2.get_parent()

                clashes_atoms.append((atom1, atom2))

                if (residue1, residue2) not in clashes_residues:
                    clashes_residues.append((residue1, residue2))

        self.clashes = len(clashes_atoms)
        self.clashes_residues = len(clashes_residues)
        return (clashes_atoms, clashes_residues)

    def get_atoms(self, chain_id=None) -> list:
        """
        Get the atoms of the structure

        Args:
            threshold: The distance threshold for a clash.

        Returns:
            A list of clashes.

        """
        if chain_id is not None:
            return [
                atom
                for chain in self.model[0]
                for atom in chain.get_atoms()
                if chain.id == chain_id
            ]
        return [atom for chain in self.model[0] for atom in chain.get_atoms()]

    def to_file(self, output_file: Union[str, Path]) -> None:
        """
        Save the cif file

        Args:
            output_file (Union[str, Path]): Path to save the cif file

        Returns:
            None
        """
        io = MMCIFIO()
        io.set_structure(self.model)

        # save creates the dictionary
        io.save(str(output_file))
        self.__atom_site_label_update(io.dic)
        self.__ligand_to_hetatm(io.dic)
        with open(output_file, "w") as f:
            io._save_dict(f)

        self.__single_to_double_quotes(output_file)

    def __single_to_double_quotes(self, file_name: Union[str, Path]) -> None:
        new_lines = []
        with open(file_name, "r") as f:
            lines = [line.rstrip() for line in f]

        for line in lines:

            single_quotes = re.compile(r"[']\w+[']{2}")
            new_lines.append(
                single_quotes.sub(lambda x: f'"{x.group()[1:-1]}"', line, count=1)
            )

        with open(file_name, "w") as f:
            f.write("\n".join(new_lines))

    def __atom_site_label_update(self, out_dict):
        atom_site_labels_asym_ids = []
        for chain_id, chain_length in self.chain_lengths(mode="all").items():
            atom_site_labels_asym_ids.extend([chain_id] * chain_length)

        assert len(out_dict["_atom_site.label_asym_id"]) == len(
            atom_site_labels_asym_ids
        ), f"Lengths must be the same, current lengths are \
{len(out_dict['_atom_site.label_asym_id'])} and {len(atom_site_labels_asym_ids)}"

        out_dict["_atom_site.label_asym_id"] = atom_site_labels_asym_ids

        return out_dict

    def __ligand_to_hetatm(self, out_dict):
        atom_site_group_pdb = []
        counter = 0
        for chain_id, chain_length in self.chain_lengths(mode="all").items():
            if self.check_ligand(chain_id):
                atom_site_group_pdb.extend(["HETATM"] * chain_length)
            else:
                atom_site_group_pdb.extend(
                    out_dict["_atom_site.group_PDB"][
                        counter : counter + chain_length  # noqa: E203
                    ]
                )

            counter += chain_length

        assert len(out_dict["_atom_site.group_PDB"]) == len(atom_site_group_pdb)

        out_dict["_atom_site.group_PDB"] = atom_site_group_pdb

        return out_dict


class ConfidenceJsonFile(FileBase):
    def __init__(self, json_file: Union[str, Path]):
        """
        Object to handle json files

        Args:
            json_file (Union[str, Path]): Path to the json file

        Attributes:
            json_file (Path): Path to the json file
            data (dict): Dictionary containing the data from the json file

        """
        super().__init__(json_file)
        self.data = self.load_json_file()

    def load_json_file(self):
        # load the json file
        with open(self.pathway, "r") as f:
            data = json.load(f)

        return data



#!/usr/bin/env python

import json
import logging
import os

from abcfold.argparse_utils import custom_template_argpase_util
from abcfold.scripts.abc_script_utils import get_custom_template

logger = logging.getLogger("logger")


def run_custom_template(
    input_json,
    target_id,
    custom_template,
    custom_template_chain,
    output_json=None,
    to_file=True,
):
    af3_json = json.load(open(input_json))

    for sequence in af3_json["sequences"]:
        if "protein" not in sequence:
            continue

        for template in custom_template:
            if not os.path.exists(template):
                msg = f"Custom template file {template} not found"
                logger.critical(msg)
                raise FileNotFoundError()
            # Can only add templates to protein sequences, so check if there
            # are multiple protein sequences in the input json
            if (
                len([x for x in af3_json["sequences"] if "protein" in x.keys()]) > 1
                and not target_id
            ):
                msg = "Multiple sequences found in input json. \
Please specify target id so that custom template can be added to the correct sequence"
                raise ValueError(msg)

        if target_id and len(target_id) > 1:
            if (len(custom_template) != len(target_id)) or (
                len(custom_template_chain) != len(target_id)
            ):
                msg = "If providing templates for multiple targets, the number of \
target ids must match the number of custom templates and custom template chains"
                raise ValueError(msg)
            custom_templates = zip(target_id, custom_template, custom_template_chain)
        else:
            if len(custom_template) != len(custom_template_chain):
                msg = "Number of custom templates must match the number of \
custom template chains"
                raise ValueError(msg)
            # if a single target id is provided, assume all custom templates
            # are for the same target
            if target_id:
                target_ids = [target_id[0]] * len(custom_template)
            else:
                target_ids = [None] * len(custom_template)
            custom_templates = zip(target_ids, custom_template, custom_template_chain)

        for i in custom_templates:
            tid, c_tem, c_tem_chn = i
            sequence = get_custom_template(
                sequence,
                tid,
                c_tem,
                c_tem_chn,
            )

    if to_file:
        if not output_json:
            output_json = input_json

        with open(output_json, "w") as f:
            json.dump(af3_json, f)

    return af3_json


def main():
    import argparse

    parser = argparse.ArgumentParser(
        description="Add custom template to alphafold input JSON"
    )

    parser.add_argument("--input_json", help="Input alphafold3 json file")
    parser.add_argument("--output_json", help="Output alphafold3 json file")
    parser = custom_template_argpase_util(parser)

    args = parser.parse_args()

    run_custom_template(  # pragma: no cover
        args.input_json,
        args.target_id,
        args.custom_template,
        args.custom_template_chain,
        output_json=args.output_json,
        to_file=True,
    )


if __name__ == "__main__":  # pragma: no cover
    main()



import argparse

from abcfold.argparse_utils import (alphafold_argparse_util,
                                    custom_template_argpase_util,
                                    main_argpase_util, mmseqs2_argparse_util)


def test_mmseqs2_argparse_util():

    parser = argparse.ArgumentParser()
    parser = mmseqs2_argparse_util(parser)
    args = parser.parse_args(["--templates", "--num_templates", "20"])
    assert args.templates
    assert args.num_templates == 20


def test_custom_template_argpase_util():
    parser = argparse.ArgumentParser()
    parser = custom_template_argpase_util(parser)
    args = parser.parse_args(
        [
            "--target_id",
            "A",
            "--custom_template",
            "test_data/test_6BJ9_cif",
            "--custom_template_chain",
            "B",
        ]
    )
    assert args.target_id == ["A"]
    assert args.custom_template == ["test_data/test_6BJ9_cif"]
    assert args.custom_template_chain == ["B"]


def test_custom_template_argpase_util_2():
    parser = argparse.ArgumentParser()
    parser = custom_template_argpase_util(parser)
    args = parser.parse_args(
        [
            "--target_id",
            "A",
            "B",
            "--custom_template",
            "test_data/test_6BJ9_cif",
            "test_data/test_6BJ9_cif",
            "--custom_template_chain",
            "A",
            "B",
        ]
    )
    assert args.target_id == ["A", "B"]
    assert args.custom_template == ["test_data/test_6BJ9_cif",
                                    "test_data/test_6BJ9_cif"]
    assert args.custom_template_chain == ["A", "B"]


def test_af3_argparse_main(test_data):
    parser = argparse.ArgumentParser()
    parser = main_argpase_util(parser)
    parser = alphafold_argparse_util(parser)
    parser = mmseqs2_argparse_util(parser)
    args = parser.parse_args(
        [
            test_data.test_inputA_json,
            "test_data",
            "--output_json",
            "test_data/test_outputA_json",
            "--database",
            "path/to/database",
            "--model_params",
            "path/to/model_params",
            "--mmseqs2",
        ]
    )
    assert args.input_json == test_data.test_inputA_json
    assert args.output_json == "test_data/test_outputA_json"
    assert args.database_dir == "path/to/database"
    assert args.model_params == "path/to/model_params"
    assert args.mmseqs2



import logging
import subprocess

from packaging.version import Version

logger = logging.getLogger(__name__)


AF3_VERSION = "3.0.0"


def check_af3_install(interactive: bool = True) -> None:
    """
    Check if Alphafold3 is installed by running the help command

    Args:
        interactive (bool): If True, run the docker container in interactive mode

    Raises:
        subprocess.CalledProcessError: If the Alphafold3 help command returns an error

    """
    logger.debug("Checking if Alphafold3 is installed")
    cmd = generate_test_command(interactive)
    with subprocess.Popen(
        cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
    ) as p:
        _, stderr = p.communicate()
        p.wait()
        if p.returncode != 1:
            logger.error(
                "Alphafold3 is not installed, please go to \
https://github.com/google-deepmind/alphafold3 and follow install instructions"
            )

            raise subprocess.CalledProcessError(p.returncode, cmd, stderr)
    logger.info("Alphafold3 is installed")

    cmd = generate_version_command()
    with subprocess.Popen(
        cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
    ) as p:
        stdout, stderr = p.communicate()
        version = str(stdout.strip().decode("utf-8"))

        if Version(version) < Version(AF3_VERSION):
            raise ImportError(
                f"Expected AlphaFold3 version {AF3_VERSION} or later, found {version}"
            )


def generate_test_command(interactive: bool = True) -> str:
    """
    Generate the Alphafold3 help command

    Args:
        interactive (bool): If True, run the docker container in interactive mode

    Returns:
        str: The Alphafold3 help command
    """
    return f"""
    docker run {'-it' if interactive else ''} \
    alphafold3 \
    python run_alphafold.py \
    --help
"""


def generate_version_command() -> str:
    """
    Generate the Alphafold3 version command
    """

    return """docker run \
    alphafold3 \
    python -c \
    'from alphafold3.version import __version__ ; print(__version__)'
"""



import pytest  # noqa: F401

from abcfold.scripts.abc_script_utils import setup_logger


def test_logging():
    try:
        logger = setup_logger()
        assert logger is not None
    except Exception as e:
        assert False, f"Exception: {e}"
    logger.debug("Debug message")
    logger.info("Info message")
    logger.warning("Warning message")
    logger.error("Error message")
    logger.critical("Critical message")



import tempfile
from pathlib import Path

import pandas as pd
import pytest

try:
    import chai_lab  # noqa F401

    from abcfold.chai1.af3_to_chai import ChaiFasta

    run_chai1 = True

except ImportError:
    run_chai1 = False


@pytest.mark.skipif(not run_chai1, reason="chai_lab not installed")
def test_af3_to_chai(test_data):
    with tempfile.TemporaryDirectory() as temp_dir:
        chai_fasta = ChaiFasta(temp_dir)

        chai_fasta.json_to_fasta(test_data.test_inputAB_json)

        reference = """>protein|A
GMRES
>protein|B
GMRES
>protein|C
YANEN
>protein|D
(ATP)
>protein|E
(ATP)
>ligand|F
CC(=O)OC1C[NH+]2CCC1CC2
"""

        filename = Path(temp_dir) / "chai1.fasta"

        assert filename.exists()
        with open(filename, "r") as f:
            data = f.read()
        print(data)
        print(reference)
        assert data == reference


@pytest.mark.skipif(not run_chai1, reason="chai_lab not installed")
def test_af3_to_chai_rna(test_data):
    with tempfile.TemporaryDirectory() as temp_dir:
        chai_fasta = ChaiFasta(temp_dir)

        chai_fasta.json_to_fasta(test_data.test_inputRNA_json)

        reference = """>rna|A\nAGCU\n"""

        filename = Path(temp_dir) / "chai1.fasta"

        assert filename.exists()
        with open(filename, "r") as f:
            data = f.read()

        assert data == reference


@pytest.mark.skipif(not run_chai1, reason="chai_lab not installed")
def test_af3_to_chai_dna(test_data):
    with tempfile.TemporaryDirectory() as temp_dir:
        chai_fasta = ChaiFasta(temp_dir)

        chai_fasta.json_to_fasta(test_data.test_inputDNA_json)

        reference = """>dna|A\nAGCT\n>dna|B\nAGCT\n"""

        filename = Path(temp_dir) / "chai1.fasta"

        assert filename.exists()
        with open(filename, "r") as f:
            data = f.read()

        assert data == reference


@pytest.mark.skipif(not run_chai1, reason="chai_lab not installed")
def test_af3_to_chai_ligand(test_data):
    with tempfile.TemporaryDirectory() as temp_dir:
        chai_fasta = ChaiFasta(temp_dir)

        chai_fasta.json_to_fasta(test_data.test_inputLIG_json)

        reference = (
            ">protein|A\n"
            "GMRESYANENQFGFKTINSDIHKIVIVGGYGKLGGLFARYLRASGYPISILDREDWAVAESILANADVV"
            "IVSVPINLTLETIERLKPYLTENMLLADLTSVKREPLAKMLEVHTGAVLGLHPMFGADIASMAKQVVVR"
            "CDGRFPERYEWLLEQIQIWGAKIYQTNATEHDHNMTYIQALRHFSTFANGLHLSKQPINLANLLALSSP"
            "IYRLELAMIGRLFAQDAELYADIIMDKSENLAVIETLKQTYDEALTFFENNDRQGFIDAFHKVRDWFGD"
            "YSEQFLKESRQLLQQANDLKQG\n"
            ">protein|B\n"
            "GMRESYANENQFGFKTINSDIHKIVIVGGYGKLGGLFARYLRASGYPISILDREDWAVAESILANADVV"
            "IVSVPINLTLETIERLKPYLTENMLLADLTSVKREPLAKMLEVHTGAVLGLHPMFGADIASMAKQVVVR"
            "CDGRFPERYEWLLEQIQIWGAKIYQTNATEHDHNMTYIQALRHFSTFANGLHLSKQPINLANLLALSSP"
            "IYRLELAMIGRLFAQDAELYADIIMDKSENLAVIETLKQTYDEALTFFENNDRQGFIDAFHKVRDWFGD"
            "YSEQFLKESRQLLQQANDLKQG\n"
            ">protein|C\n"
            "(ATP)\n"
            ">protein|D\n"
            "(ATP)\n"
            ">ligand|E\n"
            "CC(=O)OC1C[NH+]2CCC1CC2\n"
            ">ligand|G\n"
            "CCCCCCCCCCCC(O)=O\n"
            ">ligand|H\n"
            "CCCCCCCCCCCC(O)=O\n"
        )

        filename = Path(temp_dir) / "chai1.fasta"

        assert filename.exists()
        with open(filename, "r") as f:
            data = f.read()
            print(data)

        assert data == reference


@pytest.mark.skipif(not run_chai1, reason="chai_lab not installed")
def test_chai_output_constraints(test_data):
    with tempfile.TemporaryDirectory() as temp_dir:
        chai_fasta = ChaiFasta(temp_dir)

        chai_fasta.json_to_fasta(test_data.test_inputBOND_json)

        filename = Path(temp_dir) / "chai1_constraints.csv"

        assert filename.exists()

        df = pd.read_csv(filename)

        assert list(df.columns) == [
            "chainA",
            "res_idxA",
            "chainB",
            "res_idxB",
            "connection_type",
            "confidence",
            "min_distance_angstrom",
            "max_distance_angstrom",
            "comment",
            "restraint_id",
        ]

        assert len(df) == 2

        assert df.iloc[0]["chainA"] == "A"
        assert df.iloc[0]["res_idxA"] == "C387@CA"
        assert df.iloc[0]["chainB"] == "B"
        assert df.iloc[0]["res_idxB"] == "Y101@CA"
        assert df.iloc[0]["connection_type"] == "contact"
        assert df.iloc[0]["confidence"] == 1.0
        assert df.iloc[0]["min_distance_angstrom"] == 0.0
        assert df.iloc[0]["max_distance_angstrom"] == 5.5
        assert df.iloc[0]["comment"] == "Covalent Bond"
        assert df.iloc[0]["restraint_id"] == "restraint_0"


@pytest.mark.skipif(not run_chai1, reason="chai_lab not installed")
def test_chai_output_msa(test_data):
    pytest.importorskip("chai_lab")
    with tempfile.TemporaryDirectory() as temp_dir:
        chai_fasta = ChaiFasta(temp_dir)

        chai_fasta.json_to_fasta(test_data.test_inputAmsa_json)

        filename = (
            Path(temp_dir)
            / "1573f21c6a9eb5fb0727d8661eef242d8c09a481f\
c008b01ba22596b3bf31151.aligned.pqt"
        )

        assert filename.exists()

        df = pd.read_parquet(filename)

        ref_columns = ["sequence", "source_database", "pairing_key", "comment"]
        assert list(df.columns) == ref_columns

        assert len(df) == 4256

        assert (
            df.iloc[0]["sequence"]
            == "GMRESYANENQFGFKTINSDIHKIVIVGGYGKLGGL\
FARYLRASGYPISILDREDWAVAESILANADVVIVSVPINLTLETIERLKPYLTENMLLADLTSVKREPLAKMLEVHT\
GAVLGLHPMFGADIASMAKQVVVRCDGRFPERYEWLLEQIQIWGAKIYQTNATEHDHNMTYIQALRHFSTFANGLHLS\
KQPINLANLLALSSPIYRLELAMIGRLFAQDAELYADIIMDKSENLAVIETLKQTYDEALTFFENNDRQGFIDAFHKV\
RDWFGDYSEQFLKESRQLLQQANDLKQG"
        )
        assert df.iloc[0]["source_database"] == "query"
        assert df.iloc[0]["pairing_key"] == ""
        assert df.iloc[0]["comment"] == "101"


@pytest.mark.skipif(not run_chai1, reason="chai_lab not installed")
def test_ccd_to_smiles():
    chai_fasta = ChaiFasta(".")

    smiles = chai_fasta.ccd_to_smiles("DAO")
    assert smiles == "CCCCCCCCCCCC(O)=O"

    smiles = chai_fasta.ccd_to_smiles("NOT_A_CCD")
    assert smiles is None


@pytest.mark.skipif(not run_chai1, reason="chai_lab not installed")
def test_af3_data_json_to_fasta(output_objs):
    try:
        af3_json = output_objs.af3_output.input_json
        with tempfile.TemporaryDirectory() as temp_dir:
            chai_fasta = ChaiFasta(temp_dir)
            chai_fasta.json_to_fasta(af3_json)
    except TypeError:
        assert False



import logging
from pathlib import Path
from typing import Union

from abcfold.boltz1.af3_to_boltz1 import BoltzYaml
from abcfold.output.file_handlers import (CifFile, ConfidenceJsonFile,
                                          FileTypes, ModelCount, NpzFile)
from abcfold.output.utils import Af3Pae

logger = logging.getLogger("logger")


class BoltzOutput:
    def __init__(
        self,
        boltz_output_dir: Union[str, Path],
        input_params: dict,
        name: str,
    ):
        """
        Object to process the output of an Boltz-1 run

        Args:
            boltz_output_dir (Union[str, Path]): Path to the Boltz-1 output directory
            input_params (dict): Dictionary containing the input parameters used for the
            Boltz-1 run
            name (str): Name given to the Boltz-1 run

        Attributes:
            output_dir (Path): Path to the Boltz-1 output directory
            input_params (dict): Dictionary containing the input parameters used for the
            Boltz-1 run
            name (str): Name given to the Boltz-1 run
            output (dict): Dictionary containing the processed output the contents
            of the Boltz-1 output directory. The dictionary is structured as follows:

            {
                1: {
                    "pae": NpzFile,
                    "plddt": NpzFile,
                    "pde": NpzFile,
                    "cif": CifFile,
                    "json": ConfidenceJsonFile
                },
                2: {
                    "pae": NpzFile,
                    "plddt": NpzFile,
                    "pde": NpzFile,
                    "cif": CifFile,
                    "json": ConfidenceJsonFile
                },
                ...
            }
            pae_files (list): Ordered list of NpzFile objects containing the PAE data
            plddt_files (list): Ordered list of NpzFile objects containing the PLDDT
            data
            pde_files (list):  Ordered list of NpzFile objects containing the PDE data
            cif_files (list): Ordered list of CifFile objects containing the model data
            scores_files (list): Ordered list of ConfidenceJsonFile objects containing
            the model scores

        """
        self.output_dir = Path(boltz_output_dir)
        self.input_params = input_params
        self.name = name

        if self.output_dir.name.startswith("boltz_results_"):
            self.output_dir = self.output_dir.rename(
                self.output_dir.parent / f"boltz-1_{name}"
            )
        self.yaml_input_obj = self.get_input_yaml()
        self.output = self.process_boltz_output()

        self.pae_files = [value["pae"] for value in self.output.values()]
        self.cif_files = [value["cif"] for value in self.output.values()]
        self.pae_to_af3()
        self.af3_pae_files = [value["af3_pae"] for value in self.output.values()]
        self.plddt_files = [value["plddt"] for value in self.output.values()]
        self.pde_files = [value["pde"] for value in self.output.values()]
        self.scores_files = [value["json"] for value in self.output.values()]

    def process_boltz_output(self):
        """
        Function to process the output of a Boltz-1 run
        """
        file_groups = {}
        for pathway in self.output_dir.rglob("*"):
            number = pathway.stem.split("_model_")[-1]
            if not number.isdigit():
                continue
            number = int(number)

            file_type = pathway.suffix[1:]
            if file_type == FileTypes.NPZ.value:
                file_ = NpzFile(str(pathway))
            elif file_type == FileTypes.CIF.value:
                file_ = CifFile(str(pathway), self.input_params)

            elif file_type == FileTypes.JSON.value:
                file_ = ConfidenceJsonFile(str(pathway))
            else:
                continue
            if number not in file_groups:
                file_groups[number] = [file_]
            else:
                file_groups[number].append(file_)

        model_number_file_type_file = {}
        for model_number, files in file_groups.items():
            intermediate_dict = {}
            for file_ in sorted(files, key=lambda x: x.suffix):
                if file_.pathway.stem.startswith("pae"):
                    intermediate_dict["pae"] = file_
                elif file_.pathway.stem.startswith("plddt"):
                    intermediate_dict["plddt"] = file_
                elif file_.pathway.stem.startswith("pde"):
                    intermediate_dict["pde"] = file_
                elif file_.pathway.suffix == ".cif":
                    file_.name = f"Boltz-1_{model_number}"
                    file_ = self.update_chain_labels(file_)
                    intermediate_dict["cif"] = file_
                else:
                    intermediate_dict[file_.suffix] = file_

            model_number_file_type_file[model_number] = intermediate_dict

        model_number_file_type_file = {
            key: model_number_file_type_file[key]
            for key in sorted(model_number_file_type_file)
        }
        return model_number_file_type_file

    def add_plddt_to_cif(self):
        """
        Add the PLDDT scores to the B-factors of the CIF files as this is not done
        natively by Boltz-1

        Returns:
            None

        Raises:
            AssertionError: If the length of the PLDDT scores does not match the number
            of residues in the CIF file
        """
        for cif_file, plddt_scores in zip(self.cif_files, self.plddt_files):
            cif_file = self.update_chain_labels(cif_file)
            plddt_score = plddt_scores.data["plddt"]
            if max(plddt_score) <= 1:
                plddt_score = (plddt_score * 100).astype(float)

            chain_lengths = cif_file.chain_lengths(
                mode=ModelCount.RESIDUES, ligand_atoms=True, ptm_atoms=True
            )

            assert sum(chain_lengths.values()) == len(plddt_score), "Length mismatch"

            counter = 0
            for chain in cif_file.model[0]:

                # Boltz does ligand plddt per atom so we need to count them separately
                ligand = cif_file.check_ligand(chain)

                for residue in chain:
                    for atom in residue:
                        atom.b_iso = plddt_score[counter]
                        if ligand:
                            counter += 1
                    if not ligand:
                        counter += 1

            assert counter == len(plddt_score), "Length mismatch"
            cif_file.update()

    def pae_to_af3(self):
        """
        Convert the PAE data from Boltz-1 to the format used by Alphafold3

        Returns:
            None
        """
        for i, (pae_file, cif_file) in enumerate(zip(self.pae_files, self.cif_files)):
            pae = Af3Pae.from_boltz1(
                pae_file.data,
                cif_file,
            )

            out_name = cif_file.pathway.parent.joinpath(
                cif_file.pathway.stem + "_af3_pae.json"
            )

            pae.to_file(out_name)

            self.output[i]["af3_pae"] = ConfidenceJsonFile(out_name)

    def update_chain_labels(self, cif_file) -> CifFile:
        """
        Function to update the chain labels in the CIF file

        Args:
            cif_file (CifFile): CifFile object to update the chain labels for

        """
        cif_file.relabel_chains(
            self.yaml_input_obj.chain_ids, self.yaml_input_obj.id_links
        )
        return cif_file

    def get_input_yaml(self) -> BoltzYaml:
        """
        Get the input yaml file used for the Boltz-1 run

        Returns:
            BoltzYaml: Object containing the input yaml file
        """

        by = BoltzYaml(self.output_dir, create_files=False)
        by.json_to_yaml(self.input_params)

        return by



import hashlib
import json
import logging
import tempfile
from pathlib import Path
from typing import Dict, List, Optional, Union

import numpy as _  # noqa F401
import pandas as pd
import pyarrow as _  # noqa F401
import requests

from abcfold.output.atoms import VANDERWALLS

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


ATOMS_NAMES = sorted(list(VANDERWALLS.keys()), key=len, reverse=True)


class ChaiFasta:

    def __init__(self, working_dir: Union[str, Path], create_files: bool = True):
        """
        Object to convert an AlphaFold3 json to a fasta file compatible with CHAI-1

        Args:
            working_dir (Union[str, Path]): working directory to store the fasta file

        Attributes:
            working_dir (Path): working directory to store the fasta file
            fasta (Path): path to the fasta file
            constraints (Path): path to the constraints file
            msa_file (Optional[Union[str, Path]]): path to the msa file


        """
        self.working_dir = Path(working_dir)
        self.fasta = Path(working_dir) / "chai1.fasta"
        self.constraints = Path(working_dir) / "chai1_constraints.csv"
        self.msa_file: Optional[Union[str, Path]] = None
        self.__ids: List[Union[str, int]] = []
        self.__create_files = create_files

    @property
    def chain_ids(self):
        return self.__ids

    def bonded_pairs_to_file(self, bonded_pairs, fasta_data: dict):
        """
        Converts bonded pairs to a csv file compatible with CHAI-1

        Args:
            bonded_pairs (list): list of bonded pairs in AF3 format
            fasta_data (dict): dictionary with sequence data relating to IDs

        Returns:
            None
        """

        constraints_headers = [
            "chainA",
            "res_idxA",
            "chainB",
            "res_idxB",
            "connection_type",
            "confidence",
            "min_distance_angstrom",
            "max_distance_angstrom",
            "comment",
            "restraint_id",
        ]
        df = pd.DataFrame(columns=constraints_headers)

        for i, bonded_pair in enumerate(bonded_pairs):
            pair_data = []
            if (
                fasta_data[bonded_pair[0][0]] == "CCD-CODE_PLACEHOLDER"
                or fasta_data[bonded_pair[1][0]] == "CCD-CODE_PLACEHOLDER"
            ):
                logger.warning(
                    "Currently Chai-1 does not support most ccdcode ligands\
 in bonded pairs. This is something that we will be keeping in mind for future\
 updates"
                )
                continue
            if (
                fasta_data[bonded_pair[0][0]] == "LIGAND_PLACEHOLDER"
                or fasta_data[bonded_pair[1][0]] == "LIAGND_PLACEHOLDER"
            ):
                logger.warning(
                    "SMILES Ligand bonded paris are not implemented yet, please \
check back for updates"
                )
                continue

            for pair in bonded_pair:

                chain_id = pair[0]
                seq_idx = pair[1]
                pair_data.append(chain_id)
                res = fasta_data[chain_id][seq_idx - 1]

                pair_data.append(f"{res}{seq_idx}@{pair[2]}")

            row_data = [
                pair_data[0],
                pair_data[1],
                pair_data[2],
                pair_data[3],
                "contact",
                1.0,
                0.0,
                5.5,
                "Covalent Bond",
                f"restraint_{i}",
            ]
            df.loc[i] = row_data

        if not df.empty and self.__create_files:
            df.to_csv(self.constraints, sep=",", index=False)

    def msa_to_file(self, msa: str, file_path: Union[str, Path]):
        """
        Takes an msa string, converts it to pqt format and writes it to a file

        Args:
            msa (str): msa string (a3m format)
            file_path (Union[str, Path]): file path to write the msa to

        Returns:
            None
        """
        try:
            from chai_lab.data.parsing.msas.aligned_pqt import \
                merge_multi_a3m_to_aligned_dataframe
            from chai_lab.data.parsing.msas.data_source import MSADataSource
        except ImportError:

            logger.error(
                "Chai_lab didn't install correctly and the module is not available, \
please try running the job again or install chai_lab directly using 'pip \
install chai_lab'"
            )
            raise ImportError()

        # Convert msa to CHAI-1 format with additional MSA source information
        with tempfile.NamedTemporaryFile(suffix=".a3m", mode="w") as f:
            f.write(msa)
            f.flush()
            df = merge_multi_a3m_to_aligned_dataframe(
                msa_a3m_files={Path(f.name): MSADataSource.UNIPROT},
                insert_keys_for_sources="uniprot",
            )

        # Write the dataframe to a parquet file
        if not df.empty and self.__create_files:
            df.to_parquet(file_path)

    def json_to_fasta(self, json_file_or_dict: Union[dict, str, Path]):
        """
        Main function to convert an AlphaFold3 json to a fasta file
        compatible with CHAI-1

        Args:
            json_file_or_dict (Union[dict, str, Path]):
            json file or dictionary

        Returns:
            None
        """
        logger.info("Converting input json to a Chai-1 compatible fasta file")
        if isinstance(json_file_or_dict, str) or isinstance(json_file_or_dict, Path):
            with open(json_file_or_dict, "r") as f:
                json_dict = json.load(f)
        else:
            json_dict = json_file_or_dict

        fasta_data: Dict[str, str] = {}

        with open(self.fasta, "w") as f:
            for seq in json_dict["sequences"]:
                if "protein" in seq:
                    protein_str, fasta_data = self.add_protein(seq, fasta_data)
                    f.write(protein_str)
                if "rna" in seq:
                    nucleotide_str, fasta_data = self.add_nucleotide(
                        seq, "rna", fasta_data
                    )
                    f.write(nucleotide_str)
                if "dna" in seq:
                    nucleotide_str, fasta_data = self.add_nucleotide(
                        seq, "dna", fasta_data
                    )
                    f.write(nucleotide_str)
                if "ligand" in seq:
                    ligand_str = self.add_ligand(seq, fasta_data)
                    f.write(ligand_str)

        if "bondedAtomPairs" in json_dict.keys() and self.__create_files:
            if isinstance(json_dict["bondedAtomPairs"], list):
                bonded_pairs = json_dict["bondedAtomPairs"]
                self.bonded_pairs_to_file(bonded_pairs, fasta_data)

        if not self.__create_files:
            self.fasta.unlink()

        self.__ids = list(fasta_data.keys())

    def add_protein(self, seq: dict, fasta_data: dict):
        prot_id = seq["protein"]["id"]
        if isinstance(prot_id, list):
            protein_str = ""
            for i in seq["protein"]["id"]:
                temp_prot_str, fasta_data = self._add_protein(seq, i, fasta_data)
                protein_str += temp_prot_str
        else:
            protein_str, fasta_data = self._add_protein(seq, prot_id, fasta_data)
        return protein_str, fasta_data

    def _add_protein(self, seq: dict, prot_id: str, fasta_data: dict):
        protein_str = f">protein|{prot_id}\n{seq['protein']['sequence']}\n"
        sequence = seq["protein"]["sequence"]
        if "unpairedMsa" in seq["protein"].keys():
            seq_hash = hashlib.sha256(sequence.upper().encode()).hexdigest()
            pqt_path = Path(self.working_dir) / f"{seq_hash}.aligned.pqt"
            msa = seq["protein"]["unpairedMsa"]

            (
                self.msa_to_file(msa=msa, file_path=pqt_path)
                if self.__create_files
                else None
            )
        fasta_data[prot_id] = sequence

        return protein_str, fasta_data

    def add_nucleotide(self, seq: dict, seq_type: str, fasta_data: dict):
        nucl_id = seq[seq_type]["id"]

        if isinstance(nucl_id, list):
            nucleotide_str = ""
            for i in seq[seq_type]["id"]:
                temp_nucl_str, fasta_data = self._add_nucleotide(
                    seq, seq_type, i, fasta_data
                )
                nucleotide_str += temp_nucl_str
        else:
            nucleotide_str, fasta_data = self._add_nucleotide(
                seq, seq_type, nucl_id, fasta_data
            )

        return nucleotide_str, fasta_data

    def _add_nucleotide(self, seq: dict, seq_type: str, nucl_id: str, fasta_data: dict):
        nucleotide_str = f">{seq_type}|{nucl_id}\n{seq[seq_type]['sequence']}\n"
        fasta_data[nucl_id] = seq[seq_type]["sequence"]
        return nucleotide_str, fasta_data

    def ccd_to_smiles(self, ccd_id: str):

        assert isinstance(ccd_id, str), "CCD ID must be a string"
        logger.info(f"CCD code found in input: {ccd_id}")
        logger.info("Chai-1 currently only supports SMILES strings for ligands")
        logger.info("Attempting to retrieve SMILES from CCD")
        url = f"http://cactus.nci.nih.gov/chemical/structure/{ccd_id}/smiles"
        response = requests.get(url)
        if response.status_code == 200:
            ccd_data = response.text
            logger.info(f"SMILES retrieved: {ccd_data}")
            return ccd_data
        else:
            logger.warning(f"Could not retrieve SMILES for {ccd_id}")
            return None

    def add_ligand(self, seq: dict, fasta_data: dict):
        lig_id = seq["ligand"]["id"]
        ligand_str = ""
        if "ccdCodes" in seq["ligand"]:
            if isinstance(lig_id, str):
                lig_id = [lig_id]

            for lig in lig_id:
                if isinstance(seq["ligand"]["ccdCodes"], str):
                    ccd_codes = [seq["ligand"]["ccdCodes"]]
                else:
                    ccd_codes = seq["ligand"]["ccdCodes"]
                ligand_str += (
                    f">protein|{lig}\n{''.join([f'({ccd})' for ccd in ccd_codes])}\n"
                )
                fasta_data[lig] = "CCD-CODE_PLACEHOLDER"

            # if isinstance(lig_id, list):
            #     for i in lig_id:
            #         ccd_code = (
            #             seq["ligand"]["ccdCodes"][0]
            #             if isinstance(seq["ligand"]["ccdCodes"], list)
            #             else seq["ligand"]["ccdCodes"]
            #         )

            #         smile = self.ccd_to_smiles(ccd_code)
            #         if smile:
            #             ligand_str += f">ligand|{i}\n{smile}\n"
            # else:
            #     ccd_code = (
            #         seq["ligand"]["ccdCodes"][0]
            #         if isinstance(seq["ligand"]["ccdCodes"], list)
            #         else seq["ligand"]["ccdCodes"]
            #     )
            #     smile = self.ccd_to_smiles(ccd_code)
            #     if smile:
            #         ligand_str = f">ligand|{lig_id}\n{smile}\n"
        if "smiles" in seq["ligand"]:
            if isinstance(lig_id, list):
                for i in seq["ligand"]["id"]:
                    ligand_str += f">ligand|{i}\n{seq['ligand']['smiles']}\n"
                    fasta_data[i] = "SMILES_PLACEHOLDER"
            else:
                ligand_str = f">ligand|{lig_id}\n{seq['ligand']['smiles']}\n"
                fasta_data[lig_id] = "SMILES_PLACEHOLDER"

        return ligand_str

    def get_atom_name(self, atom: str) -> str:
        for name in ATOMS_NAMES:
            if atom.startswith(name):
                return name
        return ""



from distutils.command.build import build  # type: ignore

from setuptools import setup


class BuildCommand(build):
    user_options = build.user_options + [
        (
            "script-python-path=",
            None,
            "Path to Python interpreter to be included in the scripts",
        )
    ]

    def initialize_options(self):
        build.initialize_options(self)
        self.script_python_path = None

    def finalize_options(self):
        build.finalize_options(self)

    def run(self):
        global script_python_path
        script_python_path = self.script_python_path
        build.run(self)


with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    cmdclass={"build": BuildCommand},
    long_description=long_description,
    long_description_content_type="text/markdown",
)



import tempfile
from pathlib import Path

from abcfold.output.file_handlers import CifFile, ConfidenceJsonFile, NpzFile
from abcfold.output.utils import Af3Pae


def test_process_boltz_output(test_data, output_objs):
    boltz_output = output_objs.boltz_output
    assert str(boltz_output.output_dir) == str(
        Path(test_data.test_boltz_1_6BJ9_).parent.joinpath("boltz-1_6BJ9")
    )

    assert boltz_output.name == "6BJ9"

    assert 0 in boltz_output.output
    assert 1 in boltz_output.output

    assert "plddt" in boltz_output.output[0]
    assert "pae" in boltz_output.output[0]
    assert "pde" in boltz_output.output[0]
    assert "cif" in boltz_output.output[0]
    assert "json" in boltz_output.output[0]

    assert "plddt" in boltz_output.output[1]
    assert "pae" in boltz_output.output[1]
    assert "pde" in boltz_output.output[1]
    assert "cif" in boltz_output.output[1]
    assert "json" in boltz_output.output[1]

    assert all(isinstance(pae_file, NpzFile) for pae_file in boltz_output.pae_files)
    assert all(
        isinstance(plddt_file, NpzFile) for plddt_file in boltz_output.plddt_files
    )
    assert all(isinstance(pde_file, NpzFile) for pde_file in boltz_output.pde_files)
    assert all(isinstance(cif_file, CifFile) for cif_file in boltz_output.cif_files)
    assert all(
        isinstance(scores_file, ConfidenceJsonFile)
        for scores_file in boltz_output.scores_files
    )

    assert boltz_output.cif_files[0].chain_lengths() == {
        "A": 393,
        "B": 393,
        "C": 1,
        "D": 1,
    }

    # boltz_output.add_plddt_to_cif()
    with tempfile.TemporaryDirectory() as temp_dir_str:
        temp_dir = Path(temp_dir_str)
        for i, cif_file in enumerate(boltz_output.cif_files):
            cif_file.to_file(temp_dir / f"{i}.cif")
            assert (temp_dir / f"{i}.cif").exists()


def test_boltz_pae_to_af3_pae(test_data, output_objs):
    comparison_af3_output = output_objs.af3_output.af3_pae_files["seed-1"][0].data
    for pae_file, cif_file in zip(
        output_objs.boltz_output.pae_files, output_objs.boltz_output.cif_files
    ):
        pae = Af3Pae.from_boltz1(
            pae_file.data,
            cif_file,
        )

        with tempfile.TemporaryDirectory() as temp_dir_str:
            temp_dir = Path(temp_dir_str)
            pae.to_file(temp_dir / "pae.json")
            assert (temp_dir / "pae.json").exists()

        # for some reason the lengths are different for atom - realted things
        # If it isn't breaking the output page generation, then it's fine
        assert len(pae.scores["pae"]) == len(comparison_af3_output["pae"])

        assert len(pae.scores["contact_probs"]) == len(
            comparison_af3_output["contact_probs"]
        )
        assert len(pae.scores["token_chain_ids"]) == len(
            comparison_af3_output["token_chain_ids"]
        )
        assert len(pae.scores["token_res_ids"]) == len(
            comparison_af3_output["token_res_ids"]
        )



import logging
import subprocess
import sys

logger = logging.getLogger("logger")


CHAI_VERSION = "0.6.0"


def check_chai1():
    try:
        import chai_lab as _  # noqa F40

        cmd = [sys.executable, "-m", "pip", "show", "chai_lab"]
        with subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        ) as proc:
            stdout, stderr = proc.communicate()
            if proc.returncode != 0:
                raise subprocess.CalledProcessError(proc.returncode, cmd, stderr)

            version = None
            for line in stdout.decode().split("\n"):
                if line.startswith("Version:"):
                    version = line.split(":", 1)[1].strip()
                    break

            if version != CHAI_VERSION:
                raise ImportError(
                    f"Expected Chai-1 version {CHAI_VERSION}, found {version}"
                )
    except (ImportError, ModuleNotFoundError):
        try:
            import boltz1 as _  # noqa F401

            no_deps = True
        except (ImportError, ModuleNotFoundError):
            no_deps = False
        logger.info("Installing chai_lab package")
        logger.info("No dependencies will be installed") if no_deps else None
        cmd = [
            sys.executable,
            "-m",
            "pip",
            "install",
            f"chai_lab=={CHAI_VERSION}",
            "--no-cache-dir",
        ]
        cmd.append("--no-deps") if no_deps else None
        logger.info("Running %s", " ".join(cmd))
        with subprocess.Popen(
            cmd,
            stdout=sys.stdout,
            stderr=subprocess.PIPE,
        ) as proc:
            proc.wait()
            if proc.returncode != 0:
                if proc.stderr:
                    logger.error(proc.stderr.read().decode())
                raise subprocess.CalledProcessError(proc.returncode, proc.args)

        # add pytroch lightning to install



#!/usr/bin/env python

"""
usage: jsonify_scores.py [-h] [-o OUTPUT_PATH] input_path

Unpickles the output of an AlphaFold-Multimer run and converts it into a
`.json` file which can be used by PAE Viewer.

Warning: The pickle module is not secure. Only unpickle data you trust. It is
possible to construct malicious pickle data which will execute arbitrary code
during unpickling. Never unpickle data that could have come from an untrusted
source, or that could have been tampered with.

positional arguments:
  input_path            Path to the `.pickle` file containing scores as
                        generated by AlphaFold.

options:
  -h, --help            show this help message and exit
  -o OUTPUT_PATH, --output_path OUTPUT_PATH
                        Path to the output JSON file. Will fail if file
                        already exists. Saves the output in the same directory
                        as the input file by default.
"""
import argparse
import json
import pickle
import sys
from pathlib import Path

if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description=(
            "Unpickles the output of an AlphaFold-Multimer run and converts it"
            " into a `.json` file which can be used by PAE Viewer. Warning: The"
            " pickle module is not secure. Only unpickle data you trust. It is"
            " possible to construct malicious pickle data which will execute"
            " arbitrary code during unpickling. Never unpickle data that could"
            " have come from an untrusted source, or that could have been"
            " tampered with."
        )
    )

    parser.add_argument(
        'input_path',
        help=(
            "Path to the `.pickle` file containing scores as generated by"
            " AlphaFold."
        ),
    )

    parser.add_argument(
        '-o',
        '--output_path',
        help=(
            "Path to the output JSON file. Will fail if file already exists."
            " Saves the output in the same directory as the input file by"
            " default."
        ),
    )

    args = parser.parse_args()

    input_path = Path(args.input_path).resolve()
    output_path = args.output_path

    if output_path:
        output_path = Path(output_path).resolve()
    else:
        output_path = input_path.with_suffix('.json')

    if output_path.exists():
        sys.exit(f"File under output path '{output_path}' already exists!")

    with (
        open(input_path, 'rb') as input_file,
        open(output_path, 'w') as output_file,
    ):
        scores = pickle.load(input_file)

        output_scores = {
            'pae': scores['predicted_aligned_error'].tolist()
        }

        try:
            output_scores['max_pae'] = scores[
                'max_predicted_aligned_error'
            ].item()
        except KeyError:
            pass

        try:
            output_scores['plddt'] = scores['plddt'].tolist()
        except KeyError:
            pass

        try:
            output_scores['ptm'] = scores['ptm'].item()
        except KeyError:
            pass

        try:
            output_scores['iptm'] = scores['iptm'].item()
        except KeyError:
            pass

        json.dump(output_scores, output_file, ensure_ascii=False)



import json
import tempfile
from pathlib import Path

import pytest

from abcfold.scripts.add_mmseqs_msa import (MMseqs2Exception, add_msa_to_json,
                                            fetch_mmcif, run_mmseqs)


def test_add_msa_to_json(test_data):
    with open(test_data.test_inputA_json) as f:
        input_dict = json.load(f)
    add_msa_to_json(
        input_json=test_data.test_inputA_json,
        templates=test_data.test_6BJ9_cif,
        num_templates=20,
        chai_template_output=None,
        custom_template=None,
        custom_template_chain=None,
        target_id=None,
        input_params=None,
        output_json=None,
    )

    out_json = Path(test_data.test_inputA_json).parent / "inputA_mmseqs.json"

    assert out_json.exists()
    out_json.unlink()

    with tempfile.TemporaryDirectory() as tmpdir:
        tmpdir = Path(tmpdir)
        add_msa_to_json(
            input_json=test_data.test_inputA_json,
            templates=None,
            num_templates=20,
            chai_template_output=None,
            custom_template=None,
            custom_template_chain=None,
            target_id=None,
            input_params=input_dict,
            output_json=tmpdir / "output.json",
        )

        assert (tmpdir / "output.json").exists()
        with open(tmpdir / "output.json") as f:
            output_dict = json.load(f)

        assert "unpairedMsa" in output_dict["sequences"][0]["protein"]
        assert "pairedMsa" in output_dict["sequences"][0]["protein"]
        assert "templates" in output_dict["sequences"][0]["protein"]
        with pytest.raises(ValueError):
            add_msa_to_json(
                input_json=test_data.test_inputAB_json,
                templates=None,
                num_templates=20,
                chai_template_output=None,
                custom_template=[test_data.test_6BJ9_cif],
                custom_template_chain=[None],
                target_id=None,
                input_params=input_dict,
                output_json=tmpdir / "output.json",
            )
        with pytest.raises(FileNotFoundError):
            add_msa_to_json(
                input_json=test_data.test_inputAB_json,
                templates=None,
                num_templates=20,
                chai_template_output=None,
                custom_template=["road/to/nowhere"],
                custom_template_chain=[None],
                target_id=None,
                input_params=input_dict,
                output_json=tmpdir / "output.json",
            )


def test_run_mmseqs(test_data):
    # Code is taken from ColabFold and not modified so no need to test
    # but not exhaustive
    with open(test_data.test_inputA_json) as f:
        input_dict = json.load(f)

    with tempfile.TemporaryDirectory() as tmpdir:
        tmpdir = Path(tmpdir)

        sequence = input_dict["sequences"][0]
        input_sequence = sequence["protein"]["sequence"]

        run_mmseqs(
            input_sequence,
            tmpdir,
            use_templates=True,
            num_templates=20,
        )


def test_mms2_exception():
    with pytest.raises(MMseqs2Exception):
        raise MMseqs2Exception()


def test_fetch_mmcif():
    with tempfile.TemporaryDirectory() as tmpdir:
        tmpdir = Path(tmpdir)
        cif_str = fetch_mmcif("6BJ9", "A", 1, 100, tmpdir)
        assert cif_str is not None
        assert len(cif_str) > 0
        assert isinstance(cif_str, str)

        assert len(cif_str) == 110528
        cif_list = cif_str.split("\n")

        assert cif_list[0] == "data_6bj9"
        assert len(cif_list) == 1479



import json
import tempfile
from pathlib import Path

from abcfold.output.file_handlers import CifFile, ConfidenceJsonFile
from abcfold.output.utils import Af3Pae
from abcfold.scripts.abc_script_utils import (align_and_map, check_input_json,
                                              extract_sequence_from_mmcif,
                                              get_chains, get_mmcif)


def test_get_chains(test_data):
    chains = get_chains(test_data.test_6BJ9_cif)
    assert all(isinstance(chain, str) for chain in chains)
    assert len(chains) == 2
    assert "A" in chains
    assert "B" in chains


def test_extract_sequence_from_mmcif(test_data):
    with open(test_data.test_6BJ9_cif) as f:
        sequence = extract_sequence_from_mmcif(f)
    assert isinstance(sequence, str)
    assert len(sequence) == 770
    assert sequence.startswith("A")
    assert sequence.endswith("L")
    assert sequence.count("A") == 178
    assert " " not in sequence


def test_align_and_map(test_data):
    with open(test_data.test_6BJ9_cif) as f:
        template_seq = extract_sequence_from_mmcif(f)
    query_seq = template_seq[::-1]
    query_indices, template_indices = align_and_map(query_seq, template_seq)
    assert len(query_indices) == 385
    assert len(template_indices) == 385
    assert all(isinstance(i, int) for i in query_indices)
    assert all(isinstance(i, int) for i in template_indices)
    assert query_indices[0] == 3
    assert template_indices[0] == 2


def test_get_mmcif(test_data):
    cif_str = get_mmcif(test_data.test_6BJ9_cif, "6BJ9", "A", 1, 100)
    assert isinstance(cif_str, str)
    assert len(cif_str) == 110528
    cif_list = cif_str.split("\n")

    assert cif_list[0] == "data_6BJ9"
    assert len(cif_list) == 1479

    cif_str = get_mmcif(test_data.test_6BJ9_metadata_strip_cif, "6BJ9", "A", 1, 100)

    assert "_pdbx_audit_revision_history.revision_date" in cif_str

    cif_str = get_mmcif(test_data.test_1G03_cif, "1G03", "A", 1, 100)
    assert cif_str.startswith("data_1G03")


def test_check_input_json(test_data):
    input_json = test_data.test_inputAmsa_json
    input_json1 = check_input_json(input_json, use_af3_templates=True, test=True)
    input_json2 = check_input_json(input_json, use_af3_templates=False, test=True)

    assert "templates" in input_json1["sequences"][0]["protein"]
    assert input_json1["sequences"][0]["protein"]["templates"] is None

    assert "templates" in input_json2["sequences"][0]["protein"]
    assert input_json2["sequences"][0]["protein"]["templates"] == []


def test_clash_checker(test_data):
    cif_file = Path(test_data.test_boltz_1_6BJ9_).joinpath(
        "predictions", "test_mmseqs", "test_mmseqs_model_0.cif"
    )

    structure = CifFile(cif_file)
    clashes, _ = structure.check_clashes()
    assert isinstance(clashes, list)
    assert len(clashes) == 2


def test_af3_pae_reorder(test_data):
    test_cif = Path(test_data.test_alphafold3_6BJ9_).joinpath(
        "seed-1_sample-0", "model.cif"
    )
    test_pae = Path(test_data.test_alphafold3_6BJ9_).joinpath(
        "seed-1_sample-0", "confidences.json"
    )
    # /home/etk48667/dev/ABCFold/tests/test_data/alphafold3_6BJ9/6bj9_data.json
    input_params_path = Path(test_data.test_alphafold3_6BJ9_).joinpath("6bj9_data.json")

    with open(input_params_path, "r") as f:
        input_params = json.load(f)

    cif = CifFile(test_cif, input_params)
    pae = ConfidenceJsonFile(test_pae)
    pae_to_compare_tci = pae.data["token_chain_ids"][786:]

    cif.reorder_chains(["B", "A", "D", "C"])

    with tempfile.TemporaryDirectory() as temp_dir:
        cif.to_file(Path(temp_dir).joinpath("test.cif"))
        cif_file = Path(temp_dir).joinpath("test.cif")
        structure = CifFile(cif_file, input_params=input_params)
        assert [chain.id for chain in structure.get_chains()] == ["B", "A", "D", "C"]
        structure.update()
        assert [chain.id for chain in structure.get_chains()] == ["B", "A", "D", "C"]
        pae_obj = Af3Pae.from_alphafold3(pae.data, structure)

        assert (
            pae_obj.scores["token_chain_ids"][787 + 50 :]  # noqa: E203
            == pae_to_compare_tci[:51]
        )



import logging
import subprocess
import sys
from pathlib import Path
from typing import Union

logger = logging.getLogger("logger")


def run_alphafold3(
    input_json: Union[str, Path],
    output_dir: Union[str, Path],
    model_params: Union[str, Path],
    database_dir: Union[str, Path],
    interactive: bool = False,
    number_of_models: int = 5,
    num_recycles: int = 10,
) -> None:
    """
    Run Alphafold3 using the input JSON file

    Args:
        input_json (Union[str, Path]): Path to the input JSON file
        output_dir (Union[str, Path]): Path to the output directory
        model_params (Union[str, Path]): Path to the model parameters
        database_dir (Union[str, Path]): Path to the database directory
        interactive (bool): If True, run the docker container in interactive mode
        number_of_models (int): Number of models to generate

    Returns:
        None

    Raises:
        subprocess.CalledProcessError: If the Alphafold3 command returns an error

    """

    input_json = Path(input_json)
    output_dir = Path(output_dir)
    cmd = generate_af3_cmd(
        input_json=input_json,
        output_dir=output_dir,
        model_params=model_params,
        database_dir=database_dir,
        interactive=interactive,
        number_of_models=number_of_models,
        num_recycles=num_recycles,
    )

    logger.info("Running Alphafold3")
    with subprocess.Popen(
        cmd, shell=True, stdout=sys.stdout, stderr=subprocess.PIPE
    ) as p:
        _, stderr = p.communicate()
        if p.returncode != 0:
            logger.error(stderr.decode())
            raise subprocess.CalledProcessError(p.returncode, cmd, stderr)
    logger.info("Alphafold3 run complete")
    logger.info("Output files are in %s", output_dir)


def generate_af3_cmd(
    input_json: Union[str, Path],
    output_dir: Union[str, Path],
    model_params: Union[str, Path],
    database_dir: Union[str, Path],
    number_of_models: int = 10,
    num_recycles: int = 5,
    interactive: bool = False,
) -> str:
    """
    Generate the Alphafold3 command

    Args:
        input_json (Union[str, Path]): Path to the input JSON file
        output_dir (Union[str, Path]): Path to the output directory
        model_params (Union[str, Path]): Path to the model parameters
        database_dir (Union[str, Path]): Path to the database directory
        number_of_models (int): Number of models to generate
        interactive (bool): If True, run the docker container in interactive mode

    Returns:
        str: The Alphafold3 command
    """
    input_json = Path(input_json)
    output_dir = Path(output_dir)
    return f"""
    docker run {'-it' if interactive else ''} \
    --volume {input_json.parent.resolve()}:/root/af_input \
    --volume {output_dir.resolve()}:/root/af_output \
    --volume {model_params}:/root/models \
    --volume {database_dir}:/root/public_databases \
    --gpus all \
    alphafold3 \
    python run_alphafold.py \
    --json_path=/root/af_input/{input_json.name} \
    --model_dir=/root/models \
    --output_dir=/root/af_output \
    --num_diffusion_samples {number_of_models}\
    --num_recycles {num_recycles}
    """
